/* lpcboot - an LPC17xx MMC/SD boot loader inspired by HolgerBootloader2
   Based on a heavily modified version of ChaN's FatFs library

   Copyright (C) 2013  Ingo Korb <ingo@akana.de>
   All rights reserved.

   Redistribution and use in source and binary forms, with or without
   modification, are permitted provided that the following conditions
   are met:
   1. Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.
   2. Redistributions in binary form must reproduce the above copyright
      notice, this list of conditions and the following disclaimer in the
      documentation and/or other materials provided with the distribution.

   THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
   ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
   IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
   ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
   FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
   DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
   OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
   HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
   LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
   OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
   SUCH DAMAGE.


   main.c: The actual flash updater

*/

#include <string.h>
#include "config.h"
#include "crc-ccitt.h"
#include "lpc_iap.h"
#include "ff.h"
#include "spi.h"
#include "system.h"

/* block size for reading and flashing */
/* must be a power of two, not larger than a flash sector, not smaller than 256 */
#define BLOCK_SIZE 4096

typedef struct {
  uint32_t device_id;
  uint16_t version;
  uint16_t crc;
} bootinfo_t;

extern void start_app(void);

static FATFS fat;
static DIR dh;
static FILINFO finfo;
static FRESULT fr;
static FIL fd;
static bootinfo_t file_bi;
static const bootinfo_t* const rom_bi = (bootinfo_t *)(APPLICATION_OFFSET + BOOTINFO_OFFSET);
static uint8_t databuffer[BLOCK_SIZE];

static void loopdelay(unsigned int loops) {
  for (unsigned int i=0; i < loops; i++)
    asm volatile ("nop");
}

static void flash_file(void) {
  uint8_t *dest = (unsigned char *)APPLICATION_OFFSET;
  unsigned int remain = APPLICATION_LENGTH / sizeof(databuffer);
  unsigned int sector;

  /* erase all application sectors */
  sector = iap_address_to_sector(dest);
 
  if (iap_prepare(sector, LPC_SECTOR_COUNT - 1) != IR_SUCCESS)
    return;

  if (iap_erase(sector, LPC_SECTOR_COUNT - 1, CPUCLK_KHZ) != IR_SUCCESS)
    return;

  if (iap_blank_check(sector, LPC_SECTOR_COUNT - 1) != IR_SUCCESS)
    return;

  /* reopen file to reset offset */
  l_openfile(&fat, &finfo, &fd);

  /* seek to start of application */
  if (f_lseek(&fd, APPLICATION_OFFSET) != FR_OK)
    return;

  while (remain--) {
    /* toggle green LED */
    set_green_led(remain & 1);

    /* read data */
    if (f_read(&fd, databuffer, sizeof(databuffer)) != FR_OK)
      break;

    /* flash data */
    sector = iap_address_to_sector(dest);
    if (iap_prepare(sector, sector) != IR_SUCCESS)
      break;

    if (iap_ram2flash(dest, databuffer, sizeof(databuffer), CPUCLK_KHZ) != IR_SUCCESS)
      break;

    if (iap_compare(dest, databuffer, sizeof(databuffer)) != IR_SUCCESS)
      break;

    dest += sizeof(databuffer);
  }
}

// --- rewrite from here downward

static uint8_t validate_file(void) {
  unsigned int firstblock = 1;
  unsigned int remain;
  crc_t        crc;

  /* open file, can't fail */
  l_openfile(&fat, &finfo, &fd);

  /* read info block */
  if (f_lseek(&fd, BOOTINFO_OFFSET) != FR_OK)
    return 0;

  if (f_read(&fd, &file_bi, sizeof(bootinfo_t)) != FR_OK)
    return 0;

  /* check bootinfo contents */
  if (file_bi.device_id != BOOTLOADER_DEVID)
    return 0;

  if (rom_bi->device_id == BOOTLOADER_DEVID) {
    /* tests that are only appliciable if the ROM content is valid */

    if (file_bi.version == 0) {
      /* dev mode - skip flashing if identical CRC */
      if (file_bi.crc == rom_bi->crc)
        return 0;

    } else {
      /* check version */
      if (rom_bi->version != 0xffff || // allow overflow from 0xffff to lower
          file_bi.version <= rom_bi->version)
        return 0;
    }
  }

  /* seek to start of application */
  if (f_lseek(&fd, APPLICATION_OFFSET) != FR_OK)
    return 0;

  /* calculate CRC */
  remain = APPLICATION_LENGTH / BLOCK_SIZE;
  crc = crc_init();
  while (remain) {
    if (f_read(&fd, databuffer, BLOCK_SIZE) != FR_OK)
      return 0;

    remain--;

    if (firstblock) {
      /* clear the CRC in the application section */
      bootinfo_t *bufferbi = (bootinfo_t *)(databuffer + BOOTINFO_OFFSET);
      bufferbi->crc = 0;
      firstblock = 0;
    }

    crc = crc_update(crc, databuffer, BLOCK_SIZE);
  }

  if (crc != file_bi.crc)
    return 0;

  /* all tests passed */
  return 1;
}

static void try_update(void) {
  set_green_led(1);

  /* mount file system */
  fr = f_mount(0, &fat);
  if (fr != FR_OK) {
    set_green_led(0);
    return;
  }

  l_openroot(&fat, &dh);

  while ((f_readdir(&dh, &finfo) == FR_OK) && finfo.fname[0] != 0) {
    if (finfo.fsize == FLASH_SIZE) {
      /* candidate file found - validate and flash if valid */
      if (validate_file()) {
        flash_file();
        break;
      }
    }
  }

  set_green_led(0);
}

static void hardware_deinit(void) {
  system_deinit();
  spi_deinit();
  gpio_deinit();
}

static int check_app(void) {
  crc_t    crc;
  uint8_t  dummy;
  uint8_t  *ptr = (uint8_t *)APPLICATION_OFFSET;

  if (rom_bi->device_id != BOOTLOADER_DEVID)
    return 0;

  /* check in-flash CRC */
  crc = crc_init();
  crc = crc_update(crc, ptr, BOOTINFO_OFFSET);
  ptr += BOOTINFO_OFFSET;

  /* add ID and version fields */
  crc = crc_update(crc, ptr, 4+2);
  ptr += 4+2;

  /* add an uninitialized CRC field */
  dummy = 0;
  crc = crc_update(crc, &dummy, 1);
  crc = crc_update(crc, &dummy, 1);
  ptr += 2;

  /* add the remaining data */
  crc = crc_update(crc, ptr, APPLICATION_LENGTH - BOOTINFO_OFFSET - sizeof(bootinfo_t));

  if (crc == rom_bi->crc)
    return 1;
  else
    return 0;
}

static void try_start_app(void) {
  if (check_app()) {
    /* start app */
    hardware_deinit();
    start_app();

  } else {
    /* blink red LED ten times if application is corrupted */
    for (uint8_t i=0; i<10;i++) {
      set_red_led(0);
      loopdelay(2000000);
      set_red_led(1);
      loopdelay(2000000);
    }
    return;
  }
}

int main(void) {
  leds_init();
  system_speedup();
  set_red_led(1);
  set_green_led(0);

  while (1) {
    try_update();
    try_start_app();
  }
}
